from __future__ import division
from math import atan2, pi, sin, cos
import time


class Positionable(object):
    """
    Parent class for any object with a position. Defines operators {+, -, *, /,
    **} for its position with another positionable or (x, y) iterable.
    """
    def __init__(self, x=None, y=None):
        self.x = x
        self.y = y

    def __repr__(self):
        return '<%s at (%s, %s)>' % (self.__class__.__name__, self.x, self.y)

    def __str__(self):
        return repr(self)

    def __iter__(self):
        return iter(self.get_position())

    def __add__(self, other):
        ox, oy = other
        return Positionable(self.x + ox, self.y + oy)

    def __sub__(self, other):
        ox, oy = other
        return Positionable(self.x - ox, self.y - oy)

    def __mul__(self, amt):
        return Positionable(self.x * amt, self.y * amt)

    def __div__(self, amt):
        return Positionable(self.x / amt, self.y / amt)

    def __pow__(self, exp):
        return Positionable(self.x ** exp, self.y ** exp)

    def get_position(self):
        return self.x, self.y

    def set_position(self, x, y):
        self.x = x
        self.y = y

    def add_offset(self, add_x, add_y):
        self.x -= add_x
        self.y -= add_y

    def distance_to(self, other):
        """
        Calculate the Pythagorian distance from this positionable to another.
        """
        x, y = self.get_position()
        ox, oy = other
        return ((ox - x) ** 2 + (oy - y) ** 2) ** .5

    def translate(self, x, y):
        self.set_position(self.x + x, self.y + y)

    def scale(self, scale):
        self.set_position(self.x * scale, self.y * scale)

    def rotate(self, angle):
        c = cos(angle)
        s = sin(angle)
        self.set_position(c * self.x - s * self.y, s * self.x + c * self.y)


class MovingPositionable(Positionable):
    """
    Parent class for positionable objects that need movement calculations. For
    these calculations, the previous position is also saved.
    """
    def __init__(self, x=None, y=None):
        super(MovingPositionable, self).__init__(x, y)
        self.prev = Positionable(x, y)

    def set_position(self, x, y):
        """
        Set a new position and save the current position as the precious
        position. If no previous position has been set, set is to the new
        position so that the movement is zero.
        """
        if self.x is None or self.y is None:
            self.prev.set_position(x, y)
        else:
            self.prev.set_position(self.x, self.y)

        Positionable.set_position(self, x, y)

    def add_offset(self, add_x, add_y):
        Positionable.add_offset(self, add_x, add_y)
        self.prev.add_offset(add_x, add_y)

    def get_previous_position(self):
        return self.prev

    def rotation_around(self, center):
        """
        Calculate rotation of this positionable relative to a center
        positionable.
        """
        cx, cy = center
        px, py = self.prev
        prev_angle = atan2(px - cx, py - cy)
        current_angle = atan2(self.x - cx, self.y - cy)
        rotation = current_angle - prev_angle

        if rotation >= pi:
            return 2 * pi - rotation

        if rotation <= -pi:
            return -2 * pi - rotation

        return rotation

    def translation(self):
        """
        Calculate the movement relative to the last position as a vector
        positionable.
        """
        px, py = self.prev
        return Positionable(self.x - px, self.y - py)

    def translation_distance(self):
        return self.distance_to(self.prev)


class AcceleratedPositionable(MovingPositionable):
    """
    Parent class for positionable objects that need acceleration calculations.
    For these calculations, timestamps are saved for both the current and
    previous position.
    """
    def __init__(self, x=None, y=None):
        super(AcceleratedPositionable, self).__init__(x, y)
        self.prev_timestamp = self.current_timestamp = None

    def set_position(self, x, y):
        MovingPositionable.set_position(self, x, y)

        if self.current_timestamp is None:
            self.prev_timestamp = self.current_timestamp = time.time()
            return

        self.prev_timestamp = self.current_timestamp
        self.current_timestamp = time.time()

    def movement_time(self):
        return self.timestamp - self.prev_timestamp

    def acceleration(self):
        """
        Calculate the acceleration in pixels/second.
        """
        return self.movement_distance() / self.movement_time()
